print("import: 'aospy'")
import aospy

print("import: 'aospy.examples'")
import aospy.examples

print("import: 'aospy.test'")
import aospy.test

print("import: 'aospy.test.data'")
import aospy.test.data

print("import: 'aospy.test.data.objects'")
import aospy.test.data.objects

print("import: 'aospy.utils'")
import aospy.utils

